import csv
from pathlib import Path
from typing import Type, Sequence, Mapping, Union, Literal, Callable
from enum import Enum
import os
import numpy as np
import json
import time

import torch
import ezkl

from zkstats.computation import IModel



# ===================================================================================================
# ===================================================================================================

def verifier_define_calculation(
  dummy_data_path: str,
  selected_columns: list[str],
  # TODO: Here dummy_sel_data_path is redundant, but here to use process_data
  dummy_sel_data_path: str,
  verifier_model: Type[IModel],
  verifier_model_path: str,
) -> None:
  """
  Export the verifier model to an ONNX file.
  :param dummy_data_path: path to the dummy data file
  :param selected_columns: column names selected for computation
  :param dummy_sel_data_path: path to store generated preprocessed dummy data file
  :param verifier_model: the verifier model class
  :param verifier_model_path: path to store the generated verifier model file in onnx format
  """
  dummy_data_tensor_array = _process_data(dummy_data_path, selected_columns, dummy_sel_data_path)
  # export onnx file
  _export_onnx(verifier_model, dummy_data_tensor_array, verifier_model_path)


# TODO: Should only need the shape of data instead of the real dataset, since
# users (verifiers) call this function and they don't have the real data.
def create_dummy(data_path: str, dummy_data_path: str) -> None:
    """
    Create a dummy data file with randomized data based on the shape of the original data.
    """
    # Convert data file to json under the same directory but with suffix .json
    data_path: Path = Path(data_path)
    data_json_path = Path(data_path).with_suffix(DataExtension.JSON.value)

    data = json.loads(open(data_json_path, "r").read())
    # assume all columns have same number of rows
    dummy_data ={}
    for col in data:
        # not use same value for every column to prevent something weird, like singular matrix
        min_col = min(data[col])
        max_col = max(data[col])
        dummy_data[col] = np.round(np.random.uniform(min_col,max_col,len(data[col])),1).tolist()

    json.dump(dummy_data, open(dummy_data_path, 'w'))

# ===================================================================================================
# ===================================================================================================


def prover_gen_settings(
    data_path: str,
    selected_columns: list[str],
    sel_data_path: list[str],
    prover_model: Type[IModel],
    prover_model_path: str,
    scale: Union[list[int], Literal["default"]],
    # TODO: should be able to hardcode mode to "resources" or make it default?
    mode: Union[Literal["resources"], Literal["accuracy"]],
    settings_path: str,
):
    """
    Generate and calibrate settings for the given model and data.
    :param data_path: path to the data file
    :param selected_columns: column names selected for computation
    :param sel_data_path: path to store generated preprocessed data file
    :param prover_model: the prover model class
    :param prover_model_path: path to store the generated prover model file in onnx format
    :param scale: the scale to use for the computation. It's a list of integer or "default" for default scale
    :param mode: the mode to use for the computation. It's either "resources" or "accuracy"
    :param settings_path: path to store the generated settings file
    """
    data_tensor_array = _process_data(data_path, selected_columns, sel_data_path)

    # export onnx file
    _export_onnx(prover_model, data_tensor_array, prover_model_path)
    # gen + calibrate setting
    _gen_settings(sel_data_path, prover_model_path, scale, mode, settings_path)

# ===================================================================================================
# ===================================================================================================

def setup(
    model_path: str,
    compiled_model_path: str,
    settings_path: str,
    vk_path: str,
    pk_path: str,
) -> None:
  """
  Compile the verifier model and generate the verification key and public key.

  :param model_path: path to the model file in onnx format
  :param compiled_model_path: path to store the generated compiled verifier model
  :param settings_path: path to the settings file
  :param vk_path: path to store the generated verification key file
  :param pk_path: path to store the generated public key file
  """
  # compile circuit
  res = ezkl.compile_circuit(model_path, compiled_model_path, settings_path)
  assert res == True

  # srs path
  res = ezkl.get_srs(settings_path)

  # setup vk, pk param for use..... prover can use same pk or can init their own!
  print("==== setting up ezkl ====")
  start_time = time.time()
  res = ezkl.setup(
        compiled_model_path,
        vk_path,
        pk_path)
  end_time = time.time()
  time_setup = end_time -start_time
  print(f"Time setup: {time_setup} seconds")

  assert res == True
  assert os.path.isfile(vk_path)
  assert os.path.isfile(pk_path)
  assert os.path.isfile(settings_path)

# ===================================================================================================
# ===================================================================================================

def prover_gen_proof(
    prover_model_path: str,
    sel_data_path: str,
    witness_path: str,
    prover_compiled_model_path: str,
    settings_path: str,
    proof_path: str,
    pk_path: str,
) -> None:
    """
    Generate a proof for the given model and data.

    :param prover_model_path: path to the prover model file in onnx format
    :param sel_data_path: path to the preprocessed data file
    :param witness_path: path to store the generated witness file
    :param prover_compiled_model_path: path to store the generated compiled prover model
    :param settings_path: path to the settings file
    :param proof_path: path to store the generated proof file
    :param pk_path: path to the public key file
    """
    res = ezkl.compile_circuit(prover_model_path, prover_compiled_model_path, settings_path)
    assert res == True
    # now generate the witness file
    print('==== Generating Witness ====')
    witness = ezkl.gen_witness(sel_data_path, prover_compiled_model_path, witness_path)
    assert os.path.isfile(witness_path)
    # print(witness["outputs"])
    settings = json.load(open(settings_path))
    output_scale = settings['model_output_scales']
    # print("witness boolean: ", ezkl.vecu64_to_float(witness['outputs'][0][0], output_scale[0]))
    print("witness boolean: ", ezkl.felt_to_float(witness['outputs'][0][0], output_scale[0]))
    for i in range(len(witness['outputs'][1])):
      # print("witness result", i+1,":", ezkl.vecu64_to_float(witness['outputs'][1][i], output_scale[1]))
      print("witness result", i+1,":", ezkl.felt_to_float(witness['outputs'][1][i], output_scale[1]))

    # GENERATE A PROOF
    print("==== Generating Proof ====")
    start_time = time.time()
    res = ezkl.prove(
          witness_path,
          prover_compiled_model_path,
          pk_path,
          proof_path,
          "single",
      )

    print("proof: " ,res)
    end_time = time.time()
    time_gen_prf = end_time -start_time
    print(f"Time gen prf: {time_gen_prf} seconds")
    assert os.path.isfile(proof_path)


# ===================================================================================================
# ===================================================================================================

# commitment_map is a mapping[column_name, commitment_hex]
# E.g. {
#     "columns_0": "0x...",
#     ...
# }
TCommitmentMap = Mapping[str, str]
# data_commitment is a mapping[scale, mapping[column_name, commitment_hex]]
# E.g. {
#     scale_0: {
#         "columns_0": "0x...",
#         ...
#     },
#     ...
# }
TCommitmentMaps = Mapping[str, TCommitmentMap]

def verifier_verify(proof_path: str, settings_path: str, vk_path: str, selected_columns: Sequence[str], data_commitment_path: str) -> torch.Tensor:
  """
  Verify the proof and return the result.

  :param proof_path: path to the proof file
  :param settings_path: path to the settings file
  :param vk_path: path to the verification key file
  :param expected_data_commitments: expected data commitments for columns. The i-th commitment should
    be stored in `expected_data_commitments[i]`.
  """

  # 1. First check the zk proof is valid
  res = ezkl.verify(
    proof_path,
    settings_path,
    vk_path,
  )
  # TODO: change asserts to return boolean
  assert res == True

  # 2. Check if input/output are correct
  with open(settings_path) as f:
    settings = json.load(f)
  input_scales = settings['model_input_scales']
  output_scales = settings['model_output_scales']
  with open(proof_path) as f:
    proof = json.load(f)
  proof_instance = proof["instances"][0]
  inputs = proof_instance[:len(input_scales)]
  outputs = proof_instance[len(input_scales):]
  len_inputs = len(inputs)
  len_outputs = len(outputs)
  # `instances` = input commitments + params (which is 0 in our case) + output
  assert len(proof_instance) == len_inputs + len_outputs, f"lengths mismatch: {len(proof_instance)=}, {len_inputs=}, {len_outputs=}"

  # 2.1 Check input commitments
  with open(data_commitment_path) as f:
    data_commitment = json.load(f)
  # All inputs are hashed so are commitments
  assert len_inputs == len(selected_columns), f"lengths mismatch: {len_inputs=}, {len(selected_columns)=}"
  # Sanity check
  # Check each commitment is correct
  for i, (actual_commitment, column_name) in enumerate(zip(inputs, selected_columns)):
    #  actual_commitment_str = ezkl.vecu64_to_felt(actual_commitment)
     actual_commitment_str = (actual_commitment)
     input_scale = input_scales[i]
     expected_commitment = data_commitment[str(input_scale)][column_name]
     assert actual_commitment_str == expected_commitment, f"commitment mismatch: {i=}, {actual_commitment_str=}, {expected_commitment=}"

  # 2.2 Check output is correct
  # - is a tuple (is_in_error, result)
  # - is_valid is True
  # Sanity check
  is_in_error = ezkl.felt_to_float(outputs[0], output_scales[0])
  assert is_in_error == 1.0, f"result is not within error"
  result_arr = []
  for index in range(len(outputs)-1):
    result_arr.append(ezkl.felt_to_float(outputs[index+1], output_scales[1]))
  return result_arr


# ===================================================================================================
# ===================================================================================================

def generate_data_commitment(data_path: str, scales: Sequence[int], data_commitment_path: str) -> None:
  """
  Generate and store data commitment maps for different scales so that verifiers can verify
  proofs with different scales.

  :param data_path: data file path. The format must be anything defined in `DataExtension`
  :param scales: a list of scales to use for the commitments
  :param data_commitment_path: path to store the generated data commitment maps
  """

  # Convert `data_path` to json file `data_json_path`
  data_path: Path = Path(data_path)
  data_json_path = Path(data_path).with_suffix(DataExtension.JSON.value)
  _preprocess_data_file_to_json(data_path, data_json_path)

  with open(data_json_path) as f:
    data_json = json.load(f)
  data_commitments = {
    str(scale): {
      k: _get_commitment_for_column(v, scale) for k, v in data_json.items()
    } for scale in scales
  }
  with open(data_commitment_path, "w") as f:
    json.dump(data_commitments, f)


# ===================================================================================================
# Private functions
# ===================================================================================================

def _export_onnx(model: Type[IModel], data_tensor_array: list[torch.Tensor], model_loc: str) -> None:
  circuit = model()
  try:
    circuit.preprocess(data_tensor_array)
  except AttributeError:
    pass

  device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")

  # print(device)

  circuit.to(device)

  # Flips the neural net into inference mode
  circuit.eval()
  input_names = []
  # dynamic_axes = {}

  data_tensor_tuple = ()
  for i in range(len(data_tensor_array)):
    data_tensor_tuple += (data_tensor_array[i],)
    input_index = "input"+str(i+1)
    input_names.append(input_index)
  #   dynamic_axes[input_index] = {0 : 'batch_size'}
  # dynamic_axes["output"] = {0 : 'batch_size'}

  # Export the model
  torch.onnx.export(circuit,               # model being run
                      data_tensor_tuple,                   # model input (or a tuple for multiple inputs)
                      model_loc,            # where to save the model (can be a file or file-like object)
                      export_params=True,        # store the trained parameter weights inside the model file
                      opset_version=11,          # the ONNX version to export the model to
                      do_constant_folding=True,  # whether to execute constant folding for optimization
                      input_names = input_names,   # the model's input names
                      output_names = ['output'], # the model's output names
                      # dynamic_axes=dynamic_axes
                      )


# mode is either "accuracy" or "resources"
# sel_data = selected column from data that will be used for computation
def _gen_settings(
  sel_data_path: str,
  onnx_filename: str,
  scale: Union[list[int], Literal["default"]],
  mode: Union[Literal["resources"], Literal["accuracy"]],
  settings_filename: str,
) -> None:
  print("==== Generate & Calibrate Setting ====")
  # Set input to be Poseidon Hash, and param of computation graph to be public
  # Poseidon is not homomorphic additive, maybe consider Pedersens or Dory commitment.
  gip_run_args = ezkl.PyRunArgs()
  gip_run_args.input_visibility = "hashed"  # one commitment (values hashed) for each column
  gip_run_args.param_visibility = "fixed"  # no parameters shown
  gip_run_args.output_visibility = "public"  # should be `(torch.Tensor(1.0), output)`

 # generate settings
  ezkl.gen_settings(onnx_filename, settings_filename, py_run_args=gip_run_args)
  if scale =="default":
    ezkl.calibrate_settings(
    sel_data_path, onnx_filename, settings_filename, mode)
  else:
    assert isinstance(scale, list)
    ezkl.calibrate_settings(
    sel_data_path, onnx_filename, settings_filename, mode, scales = scale)

  assert os.path.exists(settings_filename)
  assert os.path.exists(sel_data_path)
  assert os.path.exists(onnx_filename)
  f_setting = open(settings_filename, "r")
  print("scale: ", scale)
  print("setting: ", f_setting.read())


def _csv_file_to_json(old_file_path: Union[Path, str], out_data_json_path: Union[Path, str],  *, delimiter: str = ",") -> None:
    data_csv_path = Path(old_file_path)
    with open(data_csv_path, 'r') as f_csv:
        reader = csv.reader(f_csv, delimiter=delimiter, strict=True)
        # Read all data from the reader to `rows`
        rows_with_column_name = tuple(reader)
    if len(rows_with_column_name) < 1:
        raise ValueError("No column names in the CSV file")
    if len(rows_with_column_name) < 2:
        raise ValueError("No data in the CSV file")
    column_names = rows_with_column_name[0]
    rows = rows_with_column_name[1:]

    columns = [
        [
            float(rows[j][i])
            for j in range(len(rows))
        ]
        for i in range(len(rows[0]))
    ]
    data = {
        column_name: column_data
        for column_name, column_data in zip(column_names, columns)
    }
    with open(out_data_json_path, "w") as f_json:
        json.dump(data, f_json)


class DataExtension(Enum):
    CSV = ".csv"
    JSON = ".json"


DATA_FORMAT_PREPROCESSING_FUNCTION: dict[DataExtension, Callable[[Union[Path, str], Path], None]] = {
    DataExtension.CSV: _csv_file_to_json,
    DataExtension.JSON: lambda old_file_path, out_data_json_path: Path(out_data_json_path).write_text(Path(old_file_path).read_text())
}

def _preprocess_data_file_to_json(data_path: Union[Path, str], out_data_json_path: Path):
    data_file_extension = DataExtension(data_path.suffix)
    preprocess_function = DATA_FORMAT_PREPROCESSING_FUNCTION[data_file_extension]
    preprocess_function(data_path, out_data_json_path)


def _process_data(
    data_path: Union[str | Path],
    col_array: list[str],
    sel_data_path: list[str],
  ) -> list[torch.Tensor]:
    data_tensor_array=[]
    sel_data = []
    data_path: Path = Path(data_path)
    # Convert data file to json under the same directory but with suffix .json
    data_json_path = Path(data_path).with_suffix(DataExtension.JSON.value)
    _preprocess_data_file_to_json(data_path, data_json_path)
    data_onefile = json.loads(open(data_json_path, "r").read())

    for col in col_array:
      data = data_onefile[col]
      data_tensor = torch.tensor(data, dtype = torch.float32)
      data_tensor_array.append(torch.reshape(data_tensor, (-1,1)))
      sel_data.append(data)
    # Serialize data into file:
    # sel_data comes from `data`
    json.dump(dict(input_data = sel_data), open(sel_data_path, 'w'))
    return data_tensor_array


def _get_commitment_for_column(column: list[float], scale: int) -> str:
  # Ref: https://github.com/zkonduit/ezkl/discussions/633
  # serialized_data = [ezkl.float_to_vecu64(x, scale) for x in column]
  serialized_data = [ezkl.float_to_felt(x, scale) for x in column]
  res_poseidon_hash = ezkl.poseidon_hash(serialized_data)[0]
  # res_hex = ezkl.vecu64_to_felt(res_poseidon_hash[0])

  return res_poseidon_hash


from abc import abstractmethod
from typing import Callable, Type, Optional, Union

import torch
from torch import nn
import json

from .ops import (
    Operation,
    Mean,
    Median,
    GeometricMean,
    HarmonicMean,
    Mode,
    PStdev,
    PVariance,
    Stdev,
    Variance,
    Covariance,
    Correlation,
    Regression,
    IsResultPrecise,
)


DEFAULT_ERROR = 0.01
MagicNumber = 99.999


class State:
    """
    State is a container for intermediate results of computation.

    Stage 1 (current_op_index is None): for every call to State (mean, median, etc.), result
        is calculated and temporarily stored in the state. Call `set_ready_for_exporting_onnx` to indicate
    Stage 2: all operations are calculated and results are ready to be used. Call `set_ready_for_exporting_onnx`
        to indicate it's ready to generate settings.
    Stage 3 (current_op_index is not None): when exporting to onnx, when the operations are called, the results and
        the conditions are popped from the state and filled in the onnx graph.
    """
    def __init__(self, error: float) -> None:
        self.ops: list[Operation] = []
        self.bools: list[Callable[[], torch.Tensor]] = []
        self.error: float = error
        # Pointer to the current operation index. If None, it's in stage 1. If not None, it's in stage 3.
        self.current_op_index: Optional[int] = None
        self.precal_witness_path: str = None
        self.precal_witness:dict = {}
        self.isProver:bool = None
        self.op_dict:dict={}

    def set_ready_for_exporting_onnx(self) -> None:
        self.current_op_index = 0

    def mean(self, x: torch.Tensor) -> torch.Tensor:
        """
        Calculate the mean of the input tensor. The behavior should conform to
        [statistics.mean](https://docs.python.org/3/library/statistics.html#statistics.mean) in Python standard library.
        """
        return self._call_op([x], Mean)

    def median(self, x: torch.Tensor) -> torch.Tensor:
        """
        Calculate the median of the input tensor. The behavior should conform to
        [statistics.median](https://docs.python.org/3/library/statistics.html#statistics.median) in Python standard library.
        """
        return self._call_op([x], Median)

    def geometric_mean(self, x: torch.Tensor) -> torch.Tensor:
        """
        Calculate the geometric mean of the input tensor. The behavior should conform to
        [statistics.geometric_mean](https://docs.python.org/3/library/statistics.html#statistics.geometric_mean) in Python standard library.
        """
        return self._call_op([x], GeometricMean)

    def harmonic_mean(self, x: torch.Tensor) -> torch.Tensor:
        """
        Calculate the harmonic mean of the input tensor. The behavior should conform to
        [statistics.harmonic_mean](https://docs.python.org/3/library/statistics.html#statistics.harmonic_mean) in Python standard library.
        """
        return self._call_op([x], HarmonicMean)

    def mode(self, x: torch.Tensor) -> torch.Tensor:
        """
        Calculate the mode of the input tensor. The behavior should conform to
        [statistics.mode](https://docs.python.org/3/library/statistics.html#statistics.mode) in Python standard library.
        """
        return self._call_op([x], Mode)

    def pstdev(self, x: torch.Tensor) -> torch.Tensor:
        """
        Calculate the population standard deviation of the input tensor. The behavior should conform to
        [statistics.pstdev](https://docs.python.org/3/library/statistics.html#statistics.pstdev) in Python standard library.
        """
        return self._call_op([x], PStdev)

    def pvariance(self, x: torch.Tensor) -> torch.Tensor:
        """
        Calculate the population variance of the input tensor. The behavior should conform to
        [statistics.pvariance](https://docs.python.org/3/library/statistics.html#statistics.pvariance) in Python standard library.
        """
        return self._call_op([x], PVariance)

    def stdev(self, x: torch.Tensor) -> torch.Tensor:
        """
        Calculate the sample standard deviation of the input tensor. The behavior should conform to
        [statistics.stdev](https://docs.python.org/3/library/statistics.html#statistics.stdev) in Python standard library.
        """
        return self._call_op([x], Stdev)

    def variance(self, x: torch.Tensor) -> torch.Tensor:
        """
        Calculate the sample variance of the input tensor. The behavior should conform to
        [statistics.variance](https://docs.python.org/3/library/statistics.html#statistics.variance) in Python standard library.
        """
        return self._call_op([x], Variance)

    def covariance(self, x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
        """
        Calculate the covariance of x and y. The behavior should conform to
        [statistics.covariance](https://docs.python.org/3/library/statistics.html#statistics.covariance) in Python standard library.
        """
        return self._call_op([x, y], Covariance)

    def correlation(self, x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
        """
        Calculate the correlation of x and y. The behavior should conform to
        [statistics.correlation](https://docs.python.org/3/library/statistics.html#statistics.correlation) in Python standard library.
        """
        return self._call_op([x, y], Correlation)

    def linear_regression(self, x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
        """
        Calculate the linear regression of x and y. The behavior should conform to
        [statistics.linear_regression](https://docs.python.org/3/library/statistics.html#statistics.linear_regression) in Python standard library.
        """
        # hence support only one x for now
        return self._call_op([x, y], Regression)

    # WHERE operation
    def where(self, _filter: torch.Tensor, x: torch.Tensor) -> torch.Tensor:
        """
        Calculate the where operation of x. The behavior should conform to `torch.where` in PyTorch.

        :param _filter: A boolean tensor serves as a filter
        :param x: A tensor to be filtered
        :return: filtered tensor
        """
        return torch.where(_filter, x, x-x+MagicNumber)

    def _call_op(self, x: list[torch.Tensor], op_type: Type[Operation]) -> Union[torch.Tensor, tuple[IsResultPrecise, torch.Tensor]]:
        if self.current_op_index is None:
            # for prover
            if self.isProver:
                # print('Prover side create')
                op = op_type.create(x, self.error)

                # Single witness aka result
                if isinstance(op,Mean) or isinstance(op,GeometricMean) or isinstance(op, HarmonicMean) or isinstance(op, Mode):
                    op_class_str =str(type(op)).split('.')[-1].split("'")[0]
                    if op_class_str not in self.op_dict:
                        self.precal_witness[op_class_str+"_0"] = [op.result.data.item()]
                        self.op_dict[op_class_str] = 1
                    else: 
                        self.precal_witness[op_class_str+"_"+str(self.op_dict[op_class_str])] = [op.result.data.item()]
                        self.op_dict[op_class_str]+=1
                elif isinstance(op, Median):
                    if 'Median' not in self.op_dict:
                        self.precal_witness['Median_0'] = [op.result.data.item(), op.lower.data.item(), op.upper.data.item()]
                        self.op_dict['Median']=1
                    else:
                        self.precal_witness['Median_'+str(self.op_dict['Median'])] = [op.result.data.item(), op.lower.data.item(), op.upper.data.item()]
                        self.op_dict['Median']+=1
                # std + variance stuffs
                elif isinstance(op, PStdev) or isinstance(op, PVariance) or isinstance(op, Stdev) or isinstance(op, Variance):
                    op_class_str =str(type(op)).split('.')[-1].split("'")[0]
                    if op_class_str not in self.op_dict:
                        self.precal_witness[op_class_str+"_0"] = [op.result.data.item(), op.data_mean.data.item()]
                        self.op_dict[op_class_str] = 1
                    else: 
                        self.precal_witness[op_class_str+"_"+str(self.op_dict[op_class_str])] = [op.result.data.item(), op.data_mean.data.item()]
                        self.op_dict[op_class_str]+=1
                elif isinstance(op, Covariance):
                    if 'Covariance' not in self.op_dict:
                        self.precal_witness['Covariance_0'] = [op.result.data.item(), op.x_mean.data.item(), op.y_mean.data.item()]
                        self.op_dict['Covariance']=1
                    else:
                        self.precal_witness['Covariance_'+str(self.op_dict['Covariance'])] = [op.result.data.item(), op.x_mean.data.item(), op.y_mean.data.item()]
                        self.op_dict['Covariance']+=1
                elif isinstance(op, Correlation):
                    if 'Correlation' not in self.op_dict:
                        self.precal_witness['Correlation_0'] = [op.result.data.item(), op.x_mean.data.item(), op.y_mean.data.item(), op.x_std.data.item(), op.y_std.data.item(), op.cov.data.item()]
                        self.op_dict['Correlation']=1
                    else:
                        self.precal_witness['Correlation_'+str(self.op_dict['Correlation'])] = [op.result.data.item(), op.x_mean.data.item(), op.y_mean.data.item(), op.x_std.data.item(), op.y_std.data.item(), op.cov.data.item()]
                        self.op_dict['Correlation']+=1
                elif isinstance(op, Regression):
                    result_array = []
                    for ele in op.result.data:
                        result_array.append(ele[0].item())
                    if 'Regression' not in self.op_dict:
                        self.precal_witness['Regression_0'] = [result_array]
                        self.op_dict['Regression']=1
                    else:
                        self.precal_witness['Regression_'+str(self.op_dict['Regression'])] = [result_array]
                        self.op_dict['Regression']+=1
                    # for ele in op.result.data[0]:
                    #     result_array.append(ele[0].item())
                    # if 'Regression' not in self.op_dict:
                    #     self.precal_witness['Regression_0'] = [result_array]
                    #     self.op_dict['Regression']=1
                    # else:
                    #     self.precal_witness['Regression_'+str(self.op_dict['Regression'])] = [result_array]
                    #     self.op_dict['Regression']+=1
            # for verifier
            else:
                # print('Verifier side create')
                precal_witness = json.loads(open(self.precal_witness_path, "r").read())
                op = op_type.create(x, self.error, precal_witness, self.op_dict)
                op_class_str =str(type(op)).split('.')[-1].split("'")[0]
                if op_class_str not in self.op_dict:
                    self.op_dict[op_class_str] = 1
                else:
                    self.op_dict[op_class_str]+=1
            self.ops.append(op)
            return op.result
        else:
            # Copy the current op index to a local variable since self.current_op_index will be incremented.
            current_op_index = self.current_op_index
            # Sanity check that current op index is not out of bound
            len_ops = len(self.ops)
            if current_op_index >= len(self.ops):
                raise Exception(f"current_op_index out of bound: {current_op_index=} >= {len_ops=}")

            op = self.ops[current_op_index]
            # Sanity check that the operation type matches the current op type
            if not isinstance(op, op_type):
                raise Exception(f"operation type mismatch: {op_type=} != {type(op)=}")

            # Increment the current op index
            self.current_op_index += 1

            # Push the ezkl condition, which is checked only in the last operation
            def is_precise() -> IsResultPrecise:
                return op.ezkl(x)
            self.bools.append(is_precise)

            # If this is the last operation, aggregate all `is_precise` in `self.bools`, and return (is_precise_aggregated, result)
            # else, return only result

            if current_op_index == len_ops - 1:
                # print('final op: ', op)
                # Sanity check for length of self.ops and self.bools
                len_bools = len(self.bools)
                if len_ops != len_bools:
                    raise Exception(f"length mismatch: {len_ops=} != {len_bools=}")
                is_precise_aggregated = torch.tensor(1.0)
                for i in range(len_bools):
                    res = self.bools[i]()
                    is_precise_aggregated = torch.logical_and(is_precise_aggregated, res)
                if self.isProver:
                    json.dump(self.precal_witness, open(self.precal_witness_path, 'w'))
                return is_precise_aggregated, op.result+(x[0]-x[0])[0][0]

            elif current_op_index > len_ops - 1:
                # Sanity check that current op index does not exceed the length of ops
                raise Exception(f"current_op_index out of bound: {current_op_index=} > {len_ops=}")
            else:
                return op.result+(x[0]-x[0])[0][0]


class IModel(nn.Module):
    @abstractmethod
    def preprocess(self, x: list[torch.Tensor]) -> None:
        ...

    @abstractmethod
    def forward(self, *x: list[torch.Tensor]) -> tuple[IsResultPrecise, torch.Tensor]:
        ...


# An computation function. Example:
# def computation(state: State, x: list[torch.Tensor]):
#     out_0 = state.median(x[0])
#     out_1 = state.median(x[1])
#     return state.mean(torch.tensor([out_0, out_1]).reshape(1,-1,1))
TComputation = Callable[[State, list[torch.Tensor]], torch.Tensor]


def computation_to_model(computation: TComputation, precal_witness_path:str, isProver:bool ,error: float = DEFAULT_ERROR ) -> tuple[State, Type[IModel]]:
    """
    Create a torch model from a `computation` function defined by user
    :param computation: A function that takes a State and a list of torch.Tensor, and returns a torch.Tensor
    :param error: The error tolerance for the computation.
    :return: A tuple of State and Model. The Model is a torch model that can be used for exporting to onnx.
    State is a container for intermediate results of computation, which can be useful when debugging.
    """
    state = State(error)
    # if it's verifier
    
    state.precal_witness_path= precal_witness_path
    state.isProver = isProver

    class Model(IModel):
        def preprocess(self, x: list[torch.Tensor]) -> None:
            computation(state, x)
            state.set_ready_for_exporting_onnx()

        def forward(self, *x: list[torch.Tensor]) -> tuple[IsResultPrecise, torch.Tensor]:
            # print('x sy: ')
            result =  computation(state, x)
            if len(result) ==1:
                return (x[0]-x[0])[0][0]+torch.tensor(1.0), result
            else:
                return result
    # print('state:: ', state.aggregate_witness_path)
    return state, Model




import json
import os
import sys
from typing import Type
import importlib.util

import click
import torch

from .core import prover_gen_proof, prover_gen_settings, setup, verifier_verify, generate_data_commitment
from .computation import computation_to_model

cwd = os.getcwd()
# TODO: Should make this configurable
output_dir = f"{cwd}/out"
os.makedirs(output_dir, exist_ok=True)
model_onnx_path = f"{output_dir}/model.onnx"
compiled_model_path = f"{output_dir}/model.compiled"
pk_path = f"{output_dir}/model.pk"
vk_path = f"{output_dir}/model.vk"
proof_path = f"{output_dir}/model.pf"
settings_path = f"{output_dir}/settings.json"
witness_path = f"{output_dir}/witness.json"
comb_data_path = f"{output_dir}/comb_data.json"
data_commitment_path = f"{output_dir}/data_commitment.json"

default_possible_scales = list(range(20))


@click.group()
def cli():
    pass


@click.command()
@click.argument('computation_path')
@click.argument('data_path')
def prove(computation_path: str, data_path: str):
    computation = load_computation(computation_path)
    _, model = computation_to_model(computation)
    generate_data_commitment(data_path, default_possible_scales, data_commitment_path)
    with open(data_commitment_path) as f:
        data_commitment = json.load(f)
    # By default select all columns
    selected_columns = list(data_commitment[str(default_possible_scales[0])].keys())
    prover_gen_settings(
        data_path,
        selected_columns,
        comb_data_path,
        model,
        model_onnx_path,
        "default",
        "resources",
        settings_path,
    )
    setup(
        model_onnx_path,
        compiled_model_path,
        settings_path,
        vk_path,
        pk_path,
    )
    print("Finished setup")
    prover_gen_proof(
        model_onnx_path,
        comb_data_path,
        witness_path,
        compiled_model_path,
        settings_path,
        proof_path,
        pk_path,
    )
    print("Finished generating proof")
    verifier_verify(proof_path, settings_path, vk_path, selected_columns, data_commitment_path)
    print("Proof path:", proof_path)
    print("Settings path:", settings_path)
    print("Verification key path:", vk_path)
    print("Commitment maps path:", data_commitment_path)


@click.command()
def verify():
    # Load commitment maps
    with open(data_commitment_path, "r") as f:
        data_commitment = json.load(f)
    # By default select all columns
    selected_columns = list(data_commitment[str(default_possible_scales[0])].keys())
    verifier_verify(proof_path, settings_path, vk_path, selected_columns, data_commitment_path)


@click.command()
@click.argument('data_path')
@click.argument('scale_str')
def commit(data_path: str, scale_str: str):
    """
    Now we just assume the data is a list of floats. We should be able to
    """
    scale = int(scale_str)
    generate_data_commitment(data_path, [scale], data_commitment_path)
    with open(data_commitment_path) as f:
        data_commitment = json.load(f)
    print("Commitment maps:", data_commitment)


def main():
    cli()


def load_computation(module_path: str) -> Type[torch.nn.Module]:
    """
    Load a model from a Python module.
    """
    # FIXME: This is unsafe since malicious code can be executed

    model_name = "computation"
    module_name = os.path.splitext(os.path.basename(module_path))[0]
    spec = importlib.util.spec_from_file_location(module_name, module_path)
    module = importlib.util.module_from_spec(spec)
    sys.modules[module_name] = module
    spec.loader.exec_module(module)

    try:
        return getattr(module, model_name)
    except AttributeError:
        raise ImportError(f"{model_name=} does not exist in {module_name=}")


# Register commands
cli.add_command(prove)
cli.add_command(verify)
cli.add_command(commit)


if __name__ == "__main__":
    main()



from abc import ABC, abstractmethod, abstractclassmethod
import statistics
from typing import Optional

import numpy as np
import torch

# boolean: either 1.0 or 0.0
IsResultPrecise = torch.Tensor
MagicNumber = 99.999


class Operation(ABC):
    def __init__(self, result: torch.Tensor, error: float):
        self.result = torch.nn.Parameter(data=result, requires_grad=False)
        self.error = error

    @abstractclassmethod
    def create(cls, x: list[torch.Tensor], error: float) -> 'Operation':
        ...

    @abstractmethod
    def ezkl(self, x: list[torch.Tensor]) -> IsResultPrecise:
        ...



class Mean(Operation):
    @classmethod
    def create(cls, x: list[torch.Tensor], error: float, precal_witness:Optional[dict] = None, op_dict:Optional[dict[str,int]] = None ) -> 'Mean':
        # support where statement, hopefully we can use 'nan' once onnx.isnan() is supported
        if precal_witness is None:
            # this is prover
            # print('provvv')
            return cls(torch.mean(x[0][x[0]!=MagicNumber]), error)
        else:
            # this is verifier
            # print('verrrr')
            if op_dict is None:
                return cls(torch.tensor(precal_witness['Mean_0'][0]), error)
            elif 'Mean' not in op_dict:
                return cls(torch.tensor(precal_witness['Mean_0'][0]), error)
            else:
                return cls(torch.tensor(precal_witness['Mean_'+str(op_dict['Mean'])][0]), error)


    def ezkl(self, x: list[torch.Tensor]) -> IsResultPrecise:
        x = x[0]
        size = torch.sum(torch.where(x!=MagicNumber, 1.0, 0.0))
        x = torch.where(x==MagicNumber, 0.0, x)
        return torch.abs(torch.sum(x)-size*self.result)<=torch.abs(self.error*self.result*size)


def to_1d(x: torch.Tensor) -> torch.Tensor:
    x_shape = x.size()
    # Only allows 1d array or [len(x), 1]
    if len(x_shape) == 1:
        return x
    elif len(x_shape) == 2 and x_shape[1] == 1:
        return x.reshape(-1)
    else:
        raise Exception(f"Unsupported shape: {x_shape=}")


class Median(Operation):
    def __init__(self, x: torch.Tensor, error: float,  precal_witness:Optional[dict] = None, op_dict:Optional[dict[str,int]] = None ):
        if precal_witness is None:
        # NOTE: To ensure `lower` and `upper` are a scalar, `x` must be a 1d array.
        # Otherwise, if `x` is a 3d array, `lower` and `upper` will be 2d array, which are not what
        # we want in our context. However, we tend to have x as a `[1, len(x), 1]`. In this case,
        # we need to flatten `x` to 1d array to get the correct `lower` and `upper`.
            x_1d = to_1d(x)
            x_1d = x_1d[x_1d!=MagicNumber]
            super().__init__(torch.tensor(np.median(x_1d)), error)
            sorted_x = np.sort(x_1d)
            len_x = len(x_1d)
            self.lower = torch.nn.Parameter(data = torch.tensor(sorted_x[int(len_x/2)-1], dtype = torch.float32), requires_grad=False)
            self.upper = torch.nn.Parameter(data = torch.tensor(sorted_x[int(len_x/2)], dtype = torch.float32), requires_grad=False)
        else:
            if op_dict is None:
                super().__init__(torch.tensor(precal_witness['Median_0'][0]), error)
                self.lower = torch.nn.Parameter(data = torch.tensor(precal_witness['Median_0'][1]), requires_grad=False)
                self.upper = torch.nn.Parameter(data = torch.tensor(precal_witness['Median_0'][2]), requires_grad=False)             
            elif 'Median' not in op_dict:
                super().__init__(torch.tensor(precal_witness['Median_0'][0]), error)
                self.lower = torch.nn.Parameter(data = torch.tensor(precal_witness['Median_0'][1]), requires_grad=False)
                self.upper = torch.nn.Parameter(data = torch.tensor(precal_witness['Median_0'][2]), requires_grad=False)
            else:
                super().__init__(torch.tensor(precal_witness['Median_'+str(op_dict['Median'])][0]), error)
                self.lower = torch.nn.Parameter(data = torch.tensor(precal_witness['Median_'+str(op_dict['Median'])][1]), requires_grad=False)
                self.upper = torch.nn.Parameter(data = torch.tensor(precal_witness['Median_'+str(op_dict['Median'])][2]), requires_grad=False)


    @classmethod
    def create(cls, x: list[torch.Tensor], error: float,  precal_witness:Optional[dict] = None, op_dict:Optional[dict[str,int]] = None ) -> 'Median':
        return cls(x[0],error, precal_witness, op_dict)

    def ezkl(self, x: list[torch.Tensor]) -> IsResultPrecise:
        x = x[0]
        old_size = x.size()[0]
        size = torch.sum(torch.where(x!=MagicNumber, 1.0, 0.0))
        min_x = torch.min(x)
        x = torch.where(x==MagicNumber,min_x-1, x)

        # since within 1%, we regard as same value
        count_less = torch.sum(torch.where(x < self.result, 1.0, 0.0))-(old_size-size)
        count_equal = torch.sum(torch.where(x==self.result, 1.0, 0.0))
        half_size = torch.floor(torch.div(size, 2))
        # print('hhhh: ', half_size)
        less_cons = count_less<half_size+size%2
        more_cons = count_less+count_equal>half_size

        # For count_equal == 0
        lower_exist = torch.sum(torch.where(x==self.lower, 1.0, 0.0))>0
        lower_cons = torch.sum(torch.where(x>self.lower, 1.0, 0.0))==half_size
        upper_exist = torch.sum(torch.where(x==self.upper, 1.0, 0.0))>0
        upper_cons = torch.sum(torch.where(x<self.upper, 1.0, 0.0))==half_size
        bound = count_less== half_size
        # 0.02 since 2*0.01
        bound_avg = (torch.abs(self.lower+self.upper-2*self.result)<=torch.abs(2*self.error*self.result))

        median_in_cons = torch.logical_and(less_cons, more_cons)
        median_out_cons = torch.logical_and(torch.logical_and(bound, bound_avg), torch.logical_and(torch.logical_and(lower_cons, upper_cons), torch.logical_and(lower_exist, upper_exist)))
        return torch.where(count_equal==0.0, median_out_cons, median_in_cons)


class GeometricMean(Operation):
    @classmethod
    def create(cls, x: list[torch.Tensor], error: float,  precal_witness:Optional[dict] = None, op_dict:Optional[dict[str,int]] = None) -> 'GeometricMean':
        if precal_witness is None:
            x_1d = to_1d(x[0])
            x_1d = x_1d[x_1d!=MagicNumber]
            result = torch.exp(torch.mean(torch.log(x_1d)))
            return cls(result, error)
        else:
            if op_dict is None:
                return cls(torch.tensor(precal_witness['GeometricMean_0'][0]), error)
            elif 'GeometricMean' not in op_dict:
                return cls(torch.tensor(precal_witness['GeometricMean_0'][0]), error)
            else:
                return cls(torch.tensor(precal_witness['GeometricMean_'+str(op_dict['GeometricMean'])][0]), error)

    def ezkl(self, x: list[torch.Tensor]) -> IsResultPrecise:
        # Assume x is [n, 1]
        x = x[0]
        size = torch.sum(torch.where(x!=MagicNumber, 1.0, 0.0))
        x = torch.where(x==MagicNumber, 1.0, x)
        return torch.abs((torch.log(self.result)*size)-torch.sum(torch.log(x)))<=size*torch.log(torch.tensor(1+self.error))


class HarmonicMean(Operation):
    @classmethod
    def create(cls, x: list[torch.Tensor], error: float, precal_witness:Optional[dict] = None, op_dict:Optional[dict[str,int]] = None) -> 'HarmonicMean':
        if precal_witness is None:
            x_1d = to_1d(x[0])
            x_1d = x_1d[x_1d!=MagicNumber]
            result = torch.div(1.0,torch.mean(torch.div(1.0, x_1d)))
            return cls(result, error)
        else:
            if op_dict is None:
                return cls(torch.tensor(precal_witness['HarmonicMean_0'][0]), error)
            elif 'HarmonicMean' not in op_dict:
                return cls(torch.tensor(precal_witness['HarmonicMean_0'][0]), error)
            else:
                return cls(torch.tensor(precal_witness['HarmonicMean_'+str(op_dict['HarmonicMean'])][0]), error)
     

    def ezkl(self, x: list[torch.Tensor]) -> IsResultPrecise:
        # Assume x is [n, 1]
        x = x[0]
        size = torch.sum(torch.where(x!=MagicNumber, 1.0, 0.0))
        return torch.abs((self.result*torch.sum(torch.where(x==MagicNumber, 0.0, torch.div(1.0, x)))) - size)<=torch.abs(self.error*size)


def mode_within(data_array: torch.Tensor, error: float) -> torch.Tensor:
    """
    Find the mode (the single most common data point) from the data_array.
    :param data_array: The data array.
    :param error: The error that allows the data point to be considered the same.
       For example, if error = 0.01, then 0.999 and 1.000 are considered the same.
    """
    max_sum_freq = 0
    mode = data_array[0]
    # print("arrrrr: ", data_array)
    # print("seetttt: ", torch.unique(data_array))
    for check_val in data_array:
        sum_freq = sum(1 for ele in data_array if abs(ele - check_val) <= abs(error * check_val))
        if sum_freq > max_sum_freq:
            mode = check_val
            max_sum_freq = sum_freq
    return mode


# TODO: Add class Mode_within , different from traditional mode
# class Mode_(Operation):
    # @classmethod
    # def create(cls, x: list[torch.Tensor], error: float) -> 'Mode':
    #     x_1d = to_1d(x[0])
    #     #  Mode has no result_error, just num_error which is the
    #     # deviation that two numbers are considered the same. (Make sense because
    #     # if some dataset has all different data, mode will be trivial if this is not the case)
    #     # This value doesn't depend on any scale, but on the dataset itself, and the intention
    #     # the evaluator. For example 0.01 means that data is counted as the same within 1% value range.

    #     # If wanting the strict definition of Mode, can just put this error to be 0
    #     result = torch.tensor(mode_within(x_1d, error))
    #     return cls(result, error)

    # def ezkl(self, x: list[torch.Tensor]) -> IsResultPrecise:
    #     # Assume x is [1, n, 1]
    #     x = x[0]
    #     size = x.size()[1]
    #     count_equal = torch.sum((torch.abs(x-self.result)<=torch.abs(self.error*self.result)).float())
    #     _result = torch.tensor([
    #         torch.sum((torch.abs(x-ele[0])<=torch.abs(self.error*ele[0])).float())<= count_equal
    #         for ele in x[0]
    #     ], dtype = torch.float32)
    #     return torch.sum(_result) == size


class Mode(Operation):
    @classmethod
    def create(cls, x: list[torch.Tensor], error: float,  precal_witness:Optional[dict] = None, op_dict:Optional[dict[str,int]] = None) -> 'Mode':
        if precal_witness is None:
            x_1d = to_1d(x[0])
            x_1d = x_1d[x_1d!=MagicNumber]
            # Here is traditional definition of Mode, can just put this num_error to be 0
            result = torch.tensor(mode_within(x_1d, 0))
            return cls(result, error)
        else:
            if op_dict is None:
                return cls(torch.tensor(precal_witness['Mode_0'][0]), error)
            elif 'Mode' not in op_dict:
                return cls(torch.tensor(precal_witness['Mode_0'][0]), error)
            else:
                return cls(torch.tensor(precal_witness['Mode_'+str(op_dict['Mode'])][0]), error)
     

    def ezkl(self, x: list[torch.Tensor]) -> IsResultPrecise:
        # Assume x is [n, 1]
        x = x[0]
        min_x = torch.min(x)
        old_size = x.size()[0]
        x = torch.where(x==MagicNumber, min_x-1, x)
        count_equal = torch.sum(torch.where(x==self.result, 1.0, 0.0))

        count_check = 0
        for ele in x:
            bool1 = torch.sum(torch.where(x==ele[0], 1.0, 0.0))<=count_equal
            bool2 = ele[0]==min_x-1
            count_check += torch.logical_or(bool1, bool2)
        return count_check ==old_size


class PStdev(Operation):
    def __init__(self, x: torch.Tensor, error: float,  precal_witness:Optional[dict] = None, op_dict:Optional[dict[str,int]] = None):
        if precal_witness is None:
            x_1d = to_1d(x)
            x_1d = x_1d[x_1d!=MagicNumber]
            self.data_mean = torch.nn.Parameter(data=torch.mean(x_1d), requires_grad=False)
            result = torch.sqrt(torch.var(x_1d, correction = 0))
            super().__init__(result, error)
        else:
            if op_dict is None:
                super().__init__(torch.tensor(precal_witness['PStdev_0'][0]), error)
                self.data_mean = torch.nn.Parameter(data = torch.tensor(precal_witness['PStdev_0'][1]), requires_grad=False)
            elif 'PStdev' not in op_dict:
                super().__init__(torch.tensor(precal_witness['PStdev_0'][0]), error)
                self.data_mean = torch.nn.Parameter(data = torch.tensor(precal_witness['PStdev_0'][1]), requires_grad=False)
            else:
                super().__init__(torch.tensor(precal_witness['PStdev_'+str(op_dict['PStdev'])][0]), error)
                self.data_mean = torch.nn.Parameter(data = torch.tensor(precal_witness['PStdev_'+str(op_dict['PStdev'])][1]), requires_grad=False)


    @classmethod
    def create(cls, x: list[torch.Tensor], error: float, precal_witness:Optional[dict] = None, op_dict:Optional[dict[str,int]] = None) -> 'PStdev':
        return cls(x[0], error, precal_witness, op_dict)

    def ezkl(self, x: list[torch.Tensor]) -> IsResultPrecise:
        x = x[0]
        x_fil_0 = torch.where(x==MagicNumber, 0.0, x)
        size = torch.sum(torch.where(x!=MagicNumber, 1.0, 0.0))
        x_mean_cons = torch.abs(torch.sum(x_fil_0)-size*(self.data_mean))<=torch.abs(self.error*self.data_mean*size)
        x_adj_mean = torch.where(x==MagicNumber, 0.0, x-self.data_mean)
        return torch.logical_and(
            torch.abs(torch.sum((x_adj_mean)*(x_adj_mean))-self.result*self.result*size)<=torch.abs(2*self.error*self.result*self.result*size),x_mean_cons
        )


class PVariance(Operation):
    def __init__(self, x: torch.Tensor, error: float,  precal_witness:Optional[dict] = None, op_dict:Optional[dict[str,int]] = None):
        if precal_witness is None:
            x_1d = to_1d(x)
            x_1d = x_1d[x_1d!=MagicNumber]
            self.data_mean = torch.nn.Parameter(data=torch.mean(x_1d), requires_grad=False)
            result = torch.var(x_1d, correction = 0)
            super().__init__(result, error)
        else:
            if op_dict is None:
                super().__init__(torch.tensor(precal_witness['PVariance_0'][0]), error)
                self.data_mean = torch.nn.Parameter(data = torch.tensor(precal_witness['PVariance_0'][1]), requires_grad=False)
            elif 'PVariance' not in op_dict:
                super().__init__(torch.tensor(precal_witness['PVariance_0'][0]), error)
                self.data_mean = torch.nn.Parameter(data = torch.tensor(precal_witness['PVariance_0'][1]), requires_grad=False)
            else:
                super().__init__(torch.tensor(precal_witness['PVariance_'+str(op_dict['PVariance'])][0]), error)
                self.data_mean = torch.nn.Parameter(data = torch.tensor(precal_witness['PVariance_'+str(op_dict['PVariance'])][1]), requires_grad=False)

    @classmethod
    def create(cls, x: list[torch.Tensor], error: float,  precal_witness:Optional[dict] = None, op_dict:Optional[dict[str,int]] = None) -> 'PVariance':
        return cls(x[0], error, precal_witness, op_dict)

    def ezkl(self, x: list[torch.Tensor]) -> IsResultPrecise:
        x = x[0]
        x_fil_0 = torch.where(x==MagicNumber, 0.0, x)
        size = torch.sum(torch.where(x!=MagicNumber, 1.0, 0.0))
        x_mean_cons = torch.abs(torch.sum(x_fil_0)-size*(self.data_mean))<=torch.abs(self.error*self.data_mean*size)
        x_adj_mean = torch.where(x==MagicNumber, 0.0, x-self.data_mean)
        return torch.logical_and(
            torch.abs(torch.sum((x_adj_mean)*(x_adj_mean))-self.result*size)<=torch.abs(self.error*self.result*size), x_mean_cons
        )



class Stdev(Operation):
    def __init__(self, x: torch.Tensor, error: float,  precal_witness:Optional[dict] = None, op_dict:Optional[dict[str,int]] = None):
        if precal_witness is None:
            x_1d = to_1d(x)
            x_1d = x_1d[x_1d!=MagicNumber]
            self.data_mean = torch.nn.Parameter(data=torch.mean(x_1d), requires_grad=False)
            result = torch.sqrt(torch.var(x_1d, correction = 1))
            super().__init__(result, error)
        else:
            if op_dict is None:
                super().__init__(torch.tensor(precal_witness['Stdev_0'][0]), error)
                self.data_mean = torch.nn.Parameter(data = torch.tensor(precal_witness['Stdev_0'][1]), requires_grad=False)
            elif 'Stdev' not in op_dict:
                super().__init__(torch.tensor(precal_witness['Stdev_0'][0]), error)
                self.data_mean = torch.nn.Parameter(data = torch.tensor(precal_witness['Stdev_0'][1]), requires_grad=False)
            else:
                super().__init__(torch.tensor(precal_witness['Stdev_'+str(op_dict['Stdev'])][0]), error)
                self.data_mean = torch.nn.Parameter(data = torch.tensor(precal_witness['Stdev_'+str(op_dict['Stdev'])][1]), requires_grad=False)


    @classmethod
    def create(cls, x: list[torch.Tensor], error: float,  precal_witness:Optional[dict] = None, op_dict:Optional[dict[str,int]] = None) -> 'Stdev':
        return cls(x[0], error, precal_witness, op_dict)

    def ezkl(self, x: list[torch.Tensor]) -> IsResultPrecise:
        x = x[0]
        x_fil_0 = torch.where(x==MagicNumber, 0.0, x)
        size = torch.sum(torch.where(x!=MagicNumber, 1.0, 0.0))
        x_mean_cons = torch.abs(torch.sum(x_fil_0)-size*(self.data_mean))<=torch.abs(self.error*self.data_mean*size)
        x_adj_mean = torch.where(x==MagicNumber, 0.0, x-self.data_mean)
        return torch.logical_and(
            torch.abs(torch.sum((x_adj_mean)*(x_adj_mean))-self.result*self.result*(size - 1))<=torch.abs(2*self.error*self.result*self.result*(size - 1)), x_mean_cons
        )


class Variance(Operation):
    def __init__(self, x: torch.Tensor, error: float,  precal_witness:Optional[dict] = None, op_dict:Optional[dict[str,int]] = None):
        if precal_witness is None:
            x_1d = to_1d(x)
            x_1d = x_1d[x_1d!=MagicNumber]
            self.data_mean = torch.nn.Parameter(data=torch.mean(x_1d), requires_grad=False)
            result = torch.var(x_1d, correction = 1)
            super().__init__(result, error)
        else:
            if op_dict is None:
                super().__init__(torch.tensor(precal_witness['Variance_0'][0]), error)
                self.data_mean = torch.nn.Parameter(data = torch.tensor(precal_witness['Variance_0'][1]), requires_grad=False)         
            elif 'Variance' not in op_dict:
                super().__init__(torch.tensor(precal_witness['Variance_0'][0]), error)
                self.data_mean = torch.nn.Parameter(data = torch.tensor(precal_witness['Variance_0'][1]), requires_grad=False)
            else:
                super().__init__(torch.tensor(precal_witness['Variance_'+str(op_dict['Variance'])][0]), error)
                self.data_mean = torch.nn.Parameter(data = torch.tensor(precal_witness['Variance_'+str(op_dict['Variance'])][1]), requires_grad=False)


    @classmethod
    def create(cls, x: list[torch.Tensor], error: float,  precal_witness:Optional[dict] = None, op_dict:Optional[dict[str,int]] = None) -> 'Variance':
        return cls(x[0], error, precal_witness, op_dict)

    def ezkl(self, x: list[torch.Tensor]) -> IsResultPrecise:
        x = x[0]
        x_fil_0 = torch.where(x==MagicNumber, 0.0, x)
        size = torch.sum(torch.where(x!=MagicNumber, 1.0, 0.0))
        x_mean_cons = torch.abs(torch.sum(x_fil_0)-size*(self.data_mean))<=torch.abs(self.error*self.data_mean*size)
        x_adj_mean = torch.where(x==MagicNumber, 0.0, x-self.data_mean)
        return torch.logical_and(
            torch.abs(torch.sum((x_adj_mean)*(x_adj_mean))-self.result*(size - 1))<=torch.abs(self.error*self.result*(size - 1)), x_mean_cons
        )




class Covariance(Operation):
    def __init__(self, x: torch.Tensor, y: torch.Tensor, error: float,  precal_witness:Optional[dict] = None, op_dict:Optional[dict[str,int]] = None):
        if precal_witness is None: 
            x_1d = to_1d(x)
            x_1d = x_1d[x_1d!=MagicNumber]
            y_1d = to_1d(y)
            y_1d = y_1d[y_1d!=MagicNumber]
            x_1d_list = x_1d.tolist()
            y_1d_list = y_1d.tolist()

            self.x_mean = torch.nn.Parameter(data=torch.tensor(statistics.mean(x_1d_list), dtype = torch.float32), requires_grad=False)
            self.y_mean = torch.nn.Parameter(data=torch.tensor(statistics.mean(y_1d_list), dtype = torch.float32), requires_grad=False)
            result = torch.tensor(statistics.covariance(x_1d_list, y_1d_list), dtype = torch.float32)

            super().__init__(result, error)
        else:
            if op_dict is None:
                super().__init__(torch.tensor(precal_witness['Covariance_0'][0]), error)
                self.x_mean = torch.nn.Parameter(data = torch.tensor(precal_witness['Covariance_0'][1]), requires_grad=False)
                self.y_mean = torch.nn.Parameter(data = torch.tensor(precal_witness['Covariance_0'][2]), requires_grad=False)  
            elif 'Covariance' not in op_dict:
                super().__init__(torch.tensor(precal_witness['Covariance_0'][0]), error)
                self.x_mean = torch.nn.Parameter(data = torch.tensor(precal_witness['Covariance_0'][1]), requires_grad=False)
                self.y_mean = torch.nn.Parameter(data = torch.tensor(precal_witness['Covariance_0'][2]), requires_grad=False)
            else:
                super().__init__(torch.tensor(precal_witness['Covariance_'+str(op_dict['Covariance'])][0]), error)
                self.x_mean = torch.nn.Parameter(data = torch.tensor(precal_witness['Covariance_'+str(op_dict['Covariance'])][1]), requires_grad=False)
                self.y_mean = torch.nn.Parameter(data = torch.tensor(precal_witness['Covariance_'+str(op_dict['Covariance'])][2]), requires_grad=False)

    @classmethod
    def create(cls, x: list[torch.Tensor], error: float,  precal_witness:Optional[dict] = None, op_dict:Optional[dict[str,int]] = None) -> 'Covariance':
        return cls(x[0], x[1], error, precal_witness, op_dict)

    def ezkl(self, args: list[torch.Tensor]) -> IsResultPrecise:
        x, y = args[0], args[1]
        x_fil_0 = torch.where(x==MagicNumber, 0.0, x)
        y_fil_0 = torch.where(y==MagicNumber, 0.0, y)
        size_x = torch.sum(torch.where(x!=MagicNumber, 1.0, 0.0))
        size_y = torch.sum(torch.where(y!=MagicNumber, 1.0, 0.0))
        x_mean_cons = torch.abs(torch.sum(x_fil_0)-size_x*(self.x_mean))<=torch.abs(self.error*self.x_mean*size_x)
        y_mean_cons = torch.abs(torch.sum(y_fil_0)-size_y*(self.y_mean))<=torch.abs(self.error*self.y_mean*size_y)
        x_adj_mean = torch.where(x==MagicNumber, 0.0, x-self.x_mean)
        y_adj_mean = torch.where(y==MagicNumber, 0.0, y-self.y_mean)

        return torch.logical_and(
            torch.logical_and(size_x==size_y,torch.logical_and(x_mean_cons,y_mean_cons)),
            torch.abs(torch.sum((x_adj_mean)*(y_adj_mean))-(size_x-1)*self.result)<=torch.abs(self.error*self.result*(size_x-1))
        )

# refer other constraints to correlation function, not put here since will be repetitive
def stdev_for_corr(x_adj_mean:torch.Tensor, size_x:torch.Tensor, x_std: torch.Tensor, error: float) -> torch.Tensor:
    return (
            torch.abs(torch.sum((x_adj_mean)*(x_adj_mean))-x_std*x_std*(size_x - 1))<=torch.abs(2*error*x_std*x_std*(size_x - 1))
        , x_std)
# refer other constraints to correlation function, not put here since will be repetitive
def covariance_for_corr(x_adj_mean: torch.Tensor,y_adj_mean: torch.Tensor,size_x:torch.Tensor, cov: torch.Tensor,  error: float) -> torch.Tensor:
        return (
            torch.abs(torch.sum((x_adj_mean)*(y_adj_mean))-(size_x-1)*cov)<=torch.abs(error*cov*(size_x-1))
        , cov)


class Correlation(Operation):
    def __init__(self, x: torch.Tensor, y: torch.Tensor, error: float, precal_witness:Optional[dict] = None, op_dict:Optional[dict[str,int]] = None):
        if precal_witness is None:
            x_1d = to_1d(x)
            x_1d = x_1d[x_1d!=MagicNumber]
            y_1d = to_1d(y)
            y_1d = y_1d[y_1d!=MagicNumber]
            x_1d_list = x_1d.tolist()
            y_1d_list = y_1d.tolist()
            self.x_mean = torch.nn.Parameter(data=torch.mean(x_1d), requires_grad=False)
            self.y_mean = torch.nn.Parameter(data=torch.mean(y_1d), requires_grad = False)
            self.x_std = torch.nn.Parameter(data=torch.sqrt(torch.var(x_1d, correction = 1)), requires_grad = False)
            self.y_std = torch.nn.Parameter(data=torch.sqrt(torch.var(y_1d, correction = 1)), requires_grad=False)
            self.cov = torch.nn.Parameter(data=torch.tensor(statistics.covariance(x_1d_list, y_1d_list), dtype = torch.float32), requires_grad=False)
            result = torch.tensor(statistics.correlation(x_1d_list, y_1d_list), dtype = torch.float32)

            super().__init__(result, error)
        else:
            if op_dict is None:
                super().__init__(torch.tensor(precal_witness['Correlation_0'][0]), error)
                self.x_mean = torch.nn.Parameter(data = torch.tensor(precal_witness['Correlation_0'][1]), requires_grad=False)
                self.y_mean = torch.nn.Parameter(data = torch.tensor(precal_witness['Correlation_0'][2]), requires_grad=False)
                self.x_std = torch.nn.Parameter(data = torch.tensor(precal_witness['Correlation_0'][3]), requires_grad=False)
                self.y_std = torch.nn.Parameter(data = torch.tensor(precal_witness['Correlation_0'][4]), requires_grad=False)
                self.cov = torch.nn.Parameter(data = torch.tensor(precal_witness['Correlation_0'][5]), requires_grad=False)       
            elif 'Correlation' not in op_dict:
                super().__init__(torch.tensor(precal_witness['Correlation_0'][0]), error)
                self.x_mean = torch.nn.Parameter(data = torch.tensor(precal_witness['Correlation_0'][1]), requires_grad=False)
                self.y_mean = torch.nn.Parameter(data = torch.tensor(precal_witness['Correlation_0'][2]), requires_grad=False)
                self.x_std = torch.nn.Parameter(data = torch.tensor(precal_witness['Correlation_0'][3]), requires_grad=False)
                self.y_std = torch.nn.Parameter(data = torch.tensor(precal_witness['Correlation_0'][4]), requires_grad=False)
                self.cov = torch.nn.Parameter(data = torch.tensor(precal_witness['Correlation_0'][5]), requires_grad=False)
            else:
                super().__init__(torch.tensor(precal_witness['Correlation_'+str(op_dict['Correlation'])][0]), error)
                self.x_mean = torch.nn.Parameter(data = torch.tensor(precal_witness['Correlation_'+str(op_dict['Correlation'])][1]), requires_grad=False)
                self.y_mean = torch.nn.Parameter(data = torch.tensor(precal_witness['Correlation_'+str(op_dict['Correlation'])][2]), requires_grad=False)
                self.x_std = torch.nn.Parameter(data = torch.tensor(precal_witness['Correlation_'+str(op_dict['Correlation'])][3]), requires_grad=False)
                self.y_std = torch.nn.Parameter(data = torch.tensor(precal_witness['Correlation_'+str(op_dict['Correlation'])][4]), requires_grad=False)
                self.cov = torch.nn.Parameter(data = torch.tensor(precal_witness['Correlation_'+str(op_dict['Correlation'])][5]), requires_grad=False)


    @classmethod
    def create(cls, args: list[torch.Tensor], error: float,  precal_witness:Optional[dict] = None, op_dict:Optional[dict[str,int]] = None) -> 'Correlation':
        return cls(args[0], args[1], error, precal_witness, op_dict)

    def ezkl(self, args: list[torch.Tensor]) -> IsResultPrecise:
        x, y = args[0], args[1]
        x_fil_0 = torch.where(x==MagicNumber, 0.0, x)
        y_fil_0 = torch.where(y==MagicNumber, 0.0, y)
        size_x = torch.sum(torch.where(x!=MagicNumber, 1.0, 0.0))
        size_y = torch.sum(torch.where(y!=MagicNumber, 1.0, 0.0))
        x_mean_cons = torch.abs(torch.sum(x_fil_0)-size_x*(self.x_mean))<=torch.abs(self.error*self.x_mean*size_x)
        y_mean_cons = torch.abs(torch.sum(y_fil_0)-size_y*(self.y_mean))<=torch.abs(self.error*self.y_mean*size_y)
        x_adj_mean = torch.where(x==MagicNumber, 0.0, x-self.x_mean)
        y_adj_mean = torch.where(y==MagicNumber, 0.0, y-self.y_mean)

        miscel_cons = torch.logical_and(size_x==size_y, torch.logical_and(x_mean_cons, y_mean_cons))
        bool1, cov = covariance_for_corr(x_adj_mean,y_adj_mean,size_x, self.cov, self.error)
        bool2, x_std = stdev_for_corr( x_adj_mean, size_x, self.x_std, self.error)
        bool3, y_std = stdev_for_corr( y_adj_mean, size_y, self.y_std, self.error)
        # this is correlation constraint
        bool4 = torch.abs(cov - self.result*x_std*y_std)<=torch.abs(self.error*cov)
        return torch.logical_and(torch.logical_and(torch.logical_and(bool1, bool2),torch.logical_and(bool3, bool4)), miscel_cons)


def stacked_x(args: list[float]):
    return np.column_stack((*args, np.ones_like(args[0])))


class Regression(Operation):
    def __init__(self, xs: list[torch.Tensor], y: torch.Tensor, error: float,  precal_witness:Optional[dict] = None, op_dict:Optional[dict[str,int]] = None):
        if precal_witness is None:
            x_1ds = [to_1d(i) for i in xs]
            fil_x_1ds=[]
            for x_1 in x_1ds:
                fil_x_1ds.append((x_1[x_1!=MagicNumber]).tolist())
            x_1ds = fil_x_1ds

            y_1d = to_1d(y)
            y_1d = (y_1d[y_1d!=MagicNumber]).tolist()

            x_one = stacked_x(x_1ds)
            result_1d = np.matmul(np.matmul(np.linalg.inv(np.matmul(x_one.transpose(), x_one)), x_one.transpose()), y_1d)
            # result = torch.tensor(result_1d, dtype = torch.float32).reshape(1, -1, 1)
            result = torch.tensor(result_1d, dtype = torch.float32).reshape(-1,1)
            super().__init__(result, error)
            # print('result regression: ', result)
        else:
            if op_dict is None:
                # result = torch.tensor(precal_witness['Regression_0']).reshape(1,-1,1)
                result = torch.tensor(precal_witness['Regression_0']).reshape(-1,1)
            elif 'Regression' not in op_dict:
                # result = torch.tensor(precal_witness['Regression_0']).reshape(1,-1,1)
                result = torch.tensor(precal_witness['Regression_0']).reshape(-1,1)
            else:
                # result = torch.tensor(precal_witness['Regression_'+str(op_dict['Regression'])]).reshape(1,-1,1)
                result = torch.tensor(precal_witness['Regression_'+str(op_dict['Regression'])]).reshape(-1,1)

            # for ele in precal_witness['Regression']:
            #     precal_witness_arr.append(torch.tensor(ele))
            # print('resultopppp: ', result)
            super().__init__(result,error)
            

    @classmethod
    def create(cls, args: list[torch.Tensor], error: float, precal_witness:Optional[dict] = None, op_dict:Optional[dict[str,int]] = None) -> 'Regression':
        xs = args[:-1]
        y = args[-1]
        return cls(xs, y, error, precal_witness, op_dict)

    def ezkl(self, args: list[torch.Tensor]) -> IsResultPrecise:
         # infer y from the last parameter
        y = args[-1]
        y = torch.where(y==MagicNumber,0.0, y)
        x_one = torch.cat((*args[:-1], torch.ones_like(args[0])), dim = 1)
        x_one = torch.where((x_one[:,0] ==MagicNumber).unsqueeze(-1), torch.tensor([0.0]*x_one.size()[1]), x_one)
        x_t = torch.transpose(x_one, 0, 1)

        left = x_t @ x_one @ self.result - x_t @ y
        right = self.error*x_t @ y
        abs_left = torch.where(left>=0, left, -left)
        abs_right = torch.where(right>=0, right, -right)
        return torch.where(torch.sum(torch.where(abs_left<=abs_right, 1.0, 0.0))==torch.tensor(2.0), 1.0, 0.0)

